package com.blazinggames.retro.modern.search;

// Java imports
import java.awt.*;
import java.util.*;

// Spelchan Imports
import com.blazinggames.urcl.data.Config;
import com.blazinggames.urcl.ui.*;
import com.blazinggames.urcl.ui.sprite.*;
import com.blazinggames.urcl.ui.comp.*;


/**
 * Ultimate Retro Episode 4: Modern Hide and Seek
 * Copyright (C) 2001, 2003-2004 Blazing Games Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ------------------------------------------------
 * 
 * The Modern version of the Hide and Seek game. The mechanics of the game are very simple.
 * Both the player and the computer choose locations on a 2D grid and then take turns trying
 * to find the other. The first to located where their opponent is wins.
 * 
 * @author Billy D. Spelchan
 * @version OSR1.00
 */
public class HandSDisplay 
extends Display
{
// *********************
// ***** CONSTANTS *****
// *********************

    private static final Point DEF_PLACE = new Point (0,0);
    private static final Rectangle DEF_RECT = new Rectangle(0,0,0,0);

    private static final String BACKDROP = "04BACKDROP";
    private static final String PLAYFIELD = "04PLAYRECT";
    private static final String BORDERCOLOR = "04BORDERCOLOR";
    private static final String GRIDCOLOR = "04GRIDCOLOR";
    private static final String NUM_LEVELS = "04LEVELS";
    private static final String LEVEL = "04LEVEL";
    private static final String MESSAGE_POS = "04MESSAGEPOS";
    private static final String MESSAGE_HIDE = "04MSGHIDE";
    private static final String MESSAGE_SEEK = "04MSGSEEK";
    private static final String MESSAGE_AISEEK = "04MSGAISEEK";
    private static final String MESSAGE_WIN = "04MSGWIN";
    private static final String MESSAGE_LOSE = "04MSGLOSE";
    
    private static final String COLOR_LOCATION = "04CLRLOCATION";
    private static final String COLOR_PLAYER = "04CLRPLAYER";
    private static final String COLOR_OPPONENT = "04CLROPPONENT";
    private static final String FONT = "04FONT";
    private static final String BIGFONT = "04BIGFONT";
    private static final String TEXTCOLOR = "04TEXTCOLOR";
    private static final String ROUND_POS = "04ROUND";
    private static final String LASTGUESS_POS = "04LASTGUESS";
    private static final String LASTDIRECTION_POS = "04LASTDIR";

    // Sound support
    private static final String USE_SOUND = "04USESOUND";
    private static final String SOUND_IMAGE = "04SOUNDIMAGE";
    private static final String SOUND_RECT = "04SOUNDRECT";
    private static final String CLICK_SOUND = "04CLICKSOUND";
    private static final String WIN_SOUND = "04WINSOUND";
    private static final String LOSE_SOUND = "04LOSESOUND";
    
    private static final int MODE_HIDE = 0;
    private static final int MODE_SEEK = 1;
    private static final int MODE_PLAYERANIMATION = 2;
    private static final int MODE_AISEEK = 3;
    private static final int MODE_AIANIMATION = 4;
    private static final int MODE_AIANIMATIONDONE = 5;
    private static final int MODE_PREWINANIMATION = 6;
    private static final int MODE_PLAYERWIN = 7;
    private static final int MODE_AIPREWINANIMATION = 8;
    private static final int MODE_COMPUTERWIN = 9;
    
// *********************
// ***** VARIABLES *****
// *********************

    private Image _background;
    private java.util.Random _random;

    // game layout
    private Rectangle _playrect;
    private Color _borderColor;
    private Color _gridColor;
    private TextSprite _message;
    private String _messages[];

//    private TextSprite _roundSprite;
    private SLabel _roundLabel;
    private Point _lastGuess;
//    private TextSprite _lastGuessSprite;
    private SLabel _lastGuessLabel;
    private ArrowSprite _lastDirection;

    // Layout attributes
    private Color _playerLocationColor;
    private Color _playerColor;
    private Color _opponentColor;
    private Color _textColor;
    private Font _font;
    private Font _bigFont;
    
    // Levels
    private Dimension _levelSizes[];
    private Dimension _cellSize;
    
    // Game Play
    private int _level;
    private int _gameMode;
    private int _round;
    private Point _player;
    private Point _opponent;
    private ArrowSprite _playerSprite;
    private Vector _opponentPlays;
    private Vector _playerPlays;
    private boolean _downInPlayfield;
    private Point _lastHighlight;
    // ai
    Point _aiLowGuess;
    Point _aiHighGuess;

    // Sound management
    boolean _useSound;
    Toggle _soundToggle;

    // Animation
    private ArrowSprite _animatedArrow;
    private boolean _showAnimation;
    private int _animationFrame;

// *****************************************
// ***** METHODS - Initialization aids *****
// *****************************************

    public void prepareAttributes()
    {
    	_playrect = _config.getRectangleProperty(PLAYFIELD, DEF_RECT);
    	_borderColor = _config.getColorProperty(BORDERCOLOR, Color.green);
    	_gridColor = _config.getColorProperty(GRIDCOLOR, Color.green);
        _playerLocationColor = _config.getColorProperty(COLOR_LOCATION, Color.pink);
        _playerColor = _config.getColorProperty(COLOR_PLAYER, Color.red);
    	_opponentColor = _config.getColorProperty(COLOR_OPPONENT, Color.blue);
    	_textColor = _config.getColorProperty(TEXTCOLOR, Color.black);
    	_font = _config.getFontProperty(FONT, font);
    	_bigFont = _config.getFontProperty(BIGFONT, font);
        _lastHighlight = new Point(-1, -1);
    }

    private void prepareSounds()
    {
        _useSound = (_config.getIntProperty(USE_SOUND, 0) == 1);
        _res.addImage(SOUND_IMAGE, _config.getStringProperty(SOUND_IMAGE, "sound.gif"));
        Rectangle r = _config.getRectangleProperty(SOUND_RECT, DEF_RECT);
    	_soundToggle = new Toggle(_displayHandler.getContainer(),
                    _res.getImage(SOUND_IMAGE), r.x, r.y, r.width, r.height, 2);
        if (_useSound)
        {
            _res.addAudioClip(CLICK_SOUND, _config.getStringProperty(CLICK_SOUND, "silent.au"));
            _res.addAudioClip(WIN_SOUND, _config.getStringProperty(WIN_SOUND, "silent.au"));
            _res.addAudioClip(LOSE_SOUND, _config.getStringProperty(LOSE_SOUND, "silent.au"));
        }
    }

    public void prepareSprites()
    {
        Container c = _displayHandler.getContainer();
        _messages = new String[10];

    	_messages[MODE_HIDE]=_config.getStringProperty(MESSAGE_HIDE, "Hide");
    	_messages[MODE_SEEK]=_config.getStringProperty(MESSAGE_SEEK, "Seek");
    	_messages[MODE_AISEEK]=_config.getStringProperty(MESSAGE_AISEEK, "computer");
    	_messages[MODE_PLAYERWIN]=_config.getStringProperty(MESSAGE_WIN, "Win");
    	_messages[MODE_COMPUTERWIN]=_config.getStringProperty(MESSAGE_LOSE, "Lose");
        
        _messages[MODE_PLAYERANIMATION] = _messages[MODE_AISEEK];
        _messages[MODE_AIANIMATION] = _messages[MODE_AISEEK];
        _messages[MODE_AIANIMATIONDONE] = _messages[MODE_AISEEK];
        _messages[MODE_AIPREWINANIMATION] = _messages[MODE_AISEEK];
        _messages[MODE_PREWINANIMATION] = "Good guess";
        
        Point p = _config.getPointProperty(MESSAGE_POS,DEF_PLACE);
        _message = new TextSprite(c, _messages[0], p.x, p.y);
        _message.setFontColor(_textColor);
        _message.setFont(c, _font);
        _message.setSolid(false);
        _message.setVisible(true);
        
       	_player = new Point(0,0);
        _opponent = new Point(0,0);
        _lastGuess = new Point(0,0);
    	_playerSprite = new ArrowSprite(_playerLocationColor, 0, 0,0,0,0);
        
        Rectangle rect = _config.getRectangleProperty(LASTDIRECTION_POS, DEF_RECT);
        _lastDirection = new ArrowSprite(_playerColor, 0, rect.x, rect.y, rect.width, rect.height);
        _lastDirection.setSolid(false);
        _lastDirection.setVisible(true);
        
        rect = _config.getRectangleProperty(LASTGUESS_POS, DEF_RECT);
        _lastGuessLabel = new SLabel(c, "0, 0", rect);
        _lastGuessLabel.setFontColor(_textColor);
        _lastGuessLabel.setFont(c, _font);
        _lastGuessLabel.setSolid(false);
        _lastGuessLabel.setAlignment(SLabel.ALIGN_CENTER);
        _lastGuessLabel.setVisible(true);
        
        rect = _config.getRectangleProperty(ROUND_POS, DEF_RECT);
        _roundLabel = new SLabel(c, "0", rect);
        _roundLabel.setFontColor(_textColor);
        _roundLabel.setFont(c, _bigFont);
        _roundLabel.setSolid(false);
        _roundLabel.setAlignment(SLabel.ALIGN_CENTER);
        _roundLabel.setVisible(true);
        
        _animatedArrow = new ArrowSprite(Color.black, 0, 0, 0, 1, 1);
        _animatedArrow.setSolid(false);
        _animatedArrow.setVisible(true);
    	_showAnimation = false;
    }
    
    public void prepareLevels()
    {
        Point p;
        int n = _config.getIntProperty(NUM_LEVELS, 1);
        
        _levelSizes = new Dimension[n];
        for (int cntr = 0; cntr < n; ++cntr)
        {
            p = _config.getPointProperty(LEVEL + "[" + cntr + "]", null);
            if (p == null)
            {
                _levelSizes[cntr] = new Dimension(10, 10);
            }
            else
            {
                _levelSizes[cntr] = new Dimension(p.x, p.y);
            }
        }
    }


// ***********************************
// ***** METHODS - Game Handling *****
// ***********************************

    public int rand(int low, int high)
    {
        int r = high - low + 1;
        if (r == 1)
            return low;
        return (((_random.nextInt() & Integer.MAX_VALUE) % r) + low);
    }

    public void setLevel(int n)
    {
        _level = n;
        changeMode(MODE_HIDE);
        
        _opponentPlays.removeAllElements();
        _playerPlays.removeAllElements();
        
        _cellSize = new Dimension(_playrect.width / _levelSizes[_level].width,
                    _playrect.height / _levelSizes[_level].height);
        _lastHighlight.setLocation(-1,-1);
    }
    
    private void changeMode(int n)
    {
        int nx, ny;
        Container c = _displayHandler.getContainer();
        
        _displayHandler.log("ModernHandS changing mode to " + n);
        _gameMode = n;
        _message.setText(c, _messages[n]);
        switch(n)
        {
            case MODE_HIDE:
                _roundLabel.setText(c, "Hide");
                _playerSprite.setVisible(false);
                _lastGuessLabel.setVisible(false);
                _lastDirection.setVisible(false);
                nx = ((_random.nextInt() & Integer.MAX_VALUE) % _levelSizes[_level].width) + 1;
                ny = ((_random.nextInt() & Integer.MAX_VALUE) % _levelSizes[_level].height) + 1;
                _displayHandler.log("Opponent at " + nx + "," + ny);
                _opponent.setLocation(nx, ny);
                _round = 0;
                _aiLowGuess = new Point(1, 1);
                _aiHighGuess = new Point(_levelSizes[_level].width, _levelSizes[_level].height);
                break;

            case MODE_SEEK:
                ++_round;
                _roundLabel.setText(c,""+_round);
                if (_round > 1)
                {
                    _lastGuessLabel.setText(c, "" + _lastGuess.x + "," + _lastGuess.y);
                    _lastGuessLabel.setVisible(true);
                    _lastDirection.setVisible(true);
                }
                else
                {
                    _playerSprite.setVisible(true);
                }
                break;

            case MODE_PLAYERANIMATION:
                break;
                
            case MODE_AISEEK:
                if (computerGuess() == false)
                    changeMode(MODE_AIANIMATION);
                else
                    changeMode(MODE_AIPREWINANIMATION);
                break;

            case MODE_AIANIMATION:
                break;
                
            case MODE_AIANIMATIONDONE:
                changeMode(MODE_SEEK);
                break;

            case MODE_PLAYERWIN:
                _lastGuessLabel.setText(c, "WIN");
                playSound(WIN_SOUND);
                break;

            case MODE_AIPREWINANIMATION:
                break;
                
            case MODE_COMPUTERWIN:
                playSound(LOSE_SOUND);
                break;

            default:
                System.out.println("Unknown mode switch made!!!");
        }
        _displayHandler.refreshDisplay(this);
    }

    private void makeGuess(int x, int y)
    {
        _displayHandler.log("Guess " + x + "," + y);
        // first make sure guess is not a duplicate - todo

        _lastGuess.setLocation(x, y);
        // determin direction
        int direction = getDirection(x, y, _opponent.x, _opponent.y);
        _lastDirection.setState(direction);
        ArrowSprite as = new ArrowSprite(_playerColor, direction, _playrect.x + x * _cellSize.width - _cellSize.width,
        _playrect.y + y * _cellSize.height - _cellSize.height, _cellSize.width, _cellSize.height);
        as.setSolid(false);
        as.setVisible(true);
        as.setGameLocation(x, y);
        _playerPlays.addElement(as);
        
        startAnimation(x, y, direction, _playerColor);
        if (direction != 0)
            changeMode(MODE_PLAYERANIMATION);
        else
            changeMode(MODE_PREWINANIMATION);
    }

    private boolean computerGuess()
    {
        boolean validGuess;
        int nx, ny;
        ArrowSprite as;
        Point loc;

_displayHandler.log("range: (" + _aiLowGuess.x + " to " + _aiHighGuess.x + "),(" +
                                _aiLowGuess.y + " to " + _aiHighGuess.y + ")");
        
        do
        {
            nx = rand(_aiLowGuess.x, _aiHighGuess.x);
            ny = rand(_aiLowGuess.y, _aiHighGuess.y);
            validGuess = true;
            for (int cntr = 0; cntr < _opponentPlays.size(); ++cntr)
            {
                as = (ArrowSprite)_opponentPlays.elementAt(cntr);
                loc = as.getGameLocation();
                if ( (loc.x == nx) && (loc.y == ny) )
                {
                    validGuess = false;
                    _displayHandler.log("GuessFailed: " + nx + "(" + _aiLowGuess.x + " to " + _aiHighGuess.x + ")," +
                                ny + "(" + _aiLowGuess.y + " to " + _aiHighGuess.y + ")");
                    break;
                }
            }
        }
        while (validGuess == false);
        
        _displayHandler.log("Computer guesses " + nx + "," + ny);
        if (nx <= _player.x)
            _aiLowGuess.x = nx;
        if (nx >= _player.x)
            _aiHighGuess.x = nx;
        if (ny <= _player.y)
            _aiLowGuess.y = ny;
        if (ny >= _player.y)
            _aiHighGuess.y = ny;
        
        int direction = getDirection(nx, ny, _player.x, _player.y);
        as = new ArrowSprite(_opponentColor, direction, 
                    _playrect.x + nx * _cellSize.width - _cellSize.width,
                    _playrect.y + ny * _cellSize.height - _cellSize.height,
                    _cellSize.width, _cellSize.height);
        as.setVisible(true);
        as.setSolid(false);
        as.setGameLocation(nx, ny);
        _opponentPlays.addElement(as);
        startAnimation(nx, ny, direction, _opponentColor);
        return (direction == 0);
    }

    private int getDirection (int gx, int gy, int tx, int ty)
    {
        int direction = 0;
        if (gx < tx)
        {
            if (gy < ty)
                direction = ArrowSprite.STATE_SOUTHEAST;
            else if (gy > ty)
                direction = ArrowSprite.STATE_NORTHEAST;
            else
                direction = ArrowSprite.STATE_EAST;
        }
        else if (gx > tx)
        {
            if (gy < ty)
                direction = ArrowSprite.STATE_SOUTHWEST;
            else if (gy > ty)
                direction = ArrowSprite.STATE_NORTHWEST;
            else
                direction = ArrowSprite.STATE_WEST;
        }
        else
        {
            if (gy < ty)
                direction = ArrowSprite.STATE_SOUTH;
            else if (gy > ty)
                direction = ArrowSprite.STATE_NORTH;
            else
            {
                direction = ArrowSprite.STATE_LOCATION;
            }
        }

        return direction;
    }

    private void startAnimation(int x, int y, int dir, Color c)
    {
    	_animatedArrow.setBounds(
                    _playrect.x + x * _cellSize.width - _cellSize.width - 10,
                    _playrect.y + y * _cellSize.height - _cellSize.height - 10,
                    _cellSize.width + 20, _cellSize.height + 20);
        _animatedArrow.setColor(c);
        _animatedArrow.setState(dir);
    	_animationFrame = 0;
    	_showAnimation = true;
    }


// ***********************************
// ***** METHODS -  From Display *****
// ***********************************

    public boolean initialize(DisplayHandler dh)
    {
        super.initialize(dh);

        _opponentPlays = new Vector();
        _playerPlays = new Vector();
        _random = new java.util.Random();
        dh.displayEvent(this, "LOADING 25");
        _res.addImage(BACKDROP, _config.getStringProperty(BACKDROP, "back.jpg"));
    	_background = _res.getImage(BACKDROP);
        prepareAttributes();
        dh.displayEvent(this, "LOADING 50");
        prepareSprites();
        dh.displayEvent(this, "LOADING 75");
        prepareLevels();
        prepareSounds();
        dh.displayEvent(this, "LOADING 100");
                
        return true;
    }

    public void mouse_move(int x, int y, boolean b)
    {
        if (_soundToggle.mouse_move(x, y, b))
        {
            _displayHandler.refreshDisplay(this);
        }
        else if (_playrect.contains(x, y))
        {
            int tilex = (x - _playrect.x) / _cellSize.width;
            int tiley = (y - _playrect.y) / _cellSize.height;
            if ( (_lastHighlight.x != tilex) || (_lastHighlight.y != tiley) )
            {
                _lastHighlight.setLocation(tilex, tiley);
                _displayHandler.refreshDisplay(this);
            }
        }
        else if (_lastHighlight.x != -1)
        {
            _lastHighlight.setLocation(-1,-1);
            _displayHandler.refreshDisplay(this);
        }
    }
    
    public void mouse_down(int x, int y)
    {
        _downInPlayfield = _playrect.contains(x, y);
        if (_soundToggle.mouse_down(x, y))
        {
            _displayHandler.refreshDisplay(this);
        }
    }
    
    public void mouse_up(int x, int y)
    {
        int nx, ny;
        System.out.println("mouse up at " + x + "," + y);
        if ( (_gameMode == MODE_PLAYERWIN) || (_gameMode == MODE_COMPUTERWIN) )
        {
            _displayHandler.displayEvent(this, "HANDS_TITLE");
        }
        else if ((_playrect.contains(x, y)) && (_downInPlayfield))
        {
            playSound(CLICK_SOUND);
            nx = (x - _playrect.x) / _cellSize.width;
            ny = (y - _playrect.y) / _cellSize.height;
            if (_gameMode == MODE_HIDE)
            {
                _player.setLocation(nx+1, ny+1);
                _displayHandler.log("player choose " + (nx+1) + "," + (ny+1));
                _playerSprite.setBounds(_playrect.x + nx * _cellSize.width, 
                            _playrect.y + ny * _cellSize.height,
                            _cellSize.width, _cellSize.height);
                changeMode(MODE_SEEK);
            }
            else if (_gameMode == MODE_SEEK)
            {
                makeGuess(nx+1, ny+1);
            }
        }    
        _downInPlayfield = false;
        if (_soundToggle.mouse_up(x, y))
        {
            _displayHandler.refreshDisplay(this);
        }
    }
    
    public void tick(int ms)
    {
        if (_soundToggle.tick(ms))
        {
            _displayHandler.refreshDisplay(this);
        }
        
        if (_showAnimation == true)
        {
            ++_animationFrame;
            if (_animationFrame >= 10)
            {
                _showAnimation = false;
                changeMode(_gameMode+1);
                return;
            }
            Rectangle r = _animatedArrow.getBounds();
            _animatedArrow.setBounds(r.x+1, r.y+1, r.width-2, r.height-2);
            _displayHandler.refreshDisplay(this);
        }
    }
    
    public synchronized void render(Graphics g, Component comp)
    {
        if (_background == null)
        {
            Dimension d = comp.getSize();
            g.setColor(Color.white);
            g.fillRect(0,0, d.width, d.height);
        }
        else
        {
            g.drawImage(_background, 0, 0, comp);
        }
        
        _message.render(g, comp);
        int cntr, temp;
        ArrowSprite as;
        
        _playerSprite.render(g, comp);
        
        for (cntr = 0; cntr < _opponentPlays.size(); ++cntr)
        {
            as = (ArrowSprite)_opponentPlays.elementAt(cntr);
            as.render(g, comp);
            as = (ArrowSprite)_playerPlays.elementAt(cntr);
            as.render(g, comp);
        }
        if (_playerPlays.size() > _opponentPlays.size())
        {
            as = (ArrowSprite)_playerPlays.lastElement();
            as.render(g, comp);
        }
        g.setColor(_gridColor);
        for (cntr = 1; cntr < _levelSizes[_level].width; ++cntr)
        {
            temp = _playrect.x + ( (_playrect.width / _levelSizes[_level].width) * cntr);
            g.drawLine(temp, _playrect.y, temp, _playrect.y + _playrect.height);
        }
        for (cntr = 1; cntr < _levelSizes[_level].height; ++cntr)
        {
            temp = _playrect.y + ( (_playrect.height / _levelSizes[_level].height) * cntr);
            g.drawLine(_playrect.x, temp, _playrect.x + _playrect.width, temp);
        }
        g.setColor(_borderColor);
        g.drawRect(_playrect.x, _playrect.y, _playrect.width, _playrect.height);
        if (_lastHighlight.x >= 0)
        {
            int hx = _playrect.x + _lastHighlight.x * _cellSize.width;
            int hy = _playrect.y + _lastHighlight.y * _cellSize.height;
            g.drawRect(hx, hy, _cellSize.width, _cellSize.height);
        }
        _lastDirection.render(g, comp);
        _roundLabel.render(g, comp);
        _lastGuessLabel.render(g, comp);
        _soundToggle.render(g, comp);
        
        if (_showAnimation == true)
            _animatedArrow.render(g, comp);
    }

    /**
     * If sound is enabled gets the resource manager to play the indicated sound.
     *
     * @param key sound to play
    */
    private void playSound(String key)
    {
        if ( (_useSound) && (_soundToggle.getState() == 1) )
        {
            _res.playAudio(key);
        }
    }

}
